use crate::utility::Op;

#[derive(Debug)]
pub struct ApplicationState {
    pub columns: usize,
    col_idx: Vec<usize>,
    pub tasks: Vec<Task>,
    pub new_task_popup: bool,
    pub should_quit: bool,
    pub selected_col: usize,
    pub selected_item: usize,
}

#[derive(Debug, Clone)]
pub struct Task {
    pub title: String,
    pub notes: String,
    pub tags: Vec<String>,
}

impl Task {
    pub fn new_test_task() -> Self {
        Task {
            title: "This is a test".to_string(),
            notes: "".to_string(),
            tags: vec![],
        }
    }
}

impl ApplicationState {
    pub fn new(columns: usize) -> Self {
        ApplicationState {
            columns,
            col_idx: (0..columns + 1).map(|_| 0).collect(),
            tasks: Vec::new(),
            new_task_popup: false,
            should_quit: false,
            selected_col: 0,
            selected_item: 0,
        }
    }

    /// Get a slice containing all of the tasks associated with that column
    /// and the index of the highlighted column in that
    pub fn get_col_slice(&self, col: usize) -> Vec<(&Task, bool)> {
        (self.col_idx[col]..self.col_idx[col + 1])
            .map(|idx| (&self.tasks[idx], idx == self.selected_item))
            .collect()
    }

    /// Get number of items in a column
    pub fn col_n_items(&self, i: usize) -> usize {
        self.col_idx[i + 1] - self.col_idx[i]
    }

    /// Determine whether a column is empty
    pub fn col_is_empty(&self, col: usize) -> bool {
        self.col_idx[col] == self.col_idx[col + 1]
    }

    /// Placeholder testing function to insert a new task in a column
    pub fn insert_task(&mut self, col: usize, task: Task) {
        // Compute the index of the new item
        let idx = self.col_idx[col + 1];

        // Update the other columns
        self.col_idx[col + 1..]
            .iter_mut()
            .for_each(|idx: &mut usize| *idx += 1);

        // Now insert the new task into the vector
        self.tasks.insert(idx, task);
        self.selected_item = idx;
    }

    pub fn remove_task(&mut self, col: usize, item: usize) {
        self.tasks.remove(item);
        for start_idx in self.col_idx[col + 1..].iter_mut() {
            *start_idx -= 1;
        }
    }

    /// Update the selected column
    pub fn update_selected_column(&mut self, op: &Op) {
        self.selected_col = op.apply(self.selected_col).min(self.columns - 1);
        self.selected_item = self.col_idx[self.selected_col];
    }

    /// Update the selected column
    pub fn update_selected_item(&mut self, op: &Op) {
        if self.col_n_items(self.selected_col) > 0 {
            self.selected_item = op
                .apply(self.selected_item)
                .max(self.col_idx[self.selected_col])
                .min(self.col_idx[self.selected_col + 1] - 1)
        };
    }

    /// Move a task either between columns or within a column
    pub fn move_task(&mut self, col: usize, item: usize, op: &Op, in_col: bool) {
        if in_col {
            let new_idx = op
                .apply(item)
                .max(self.col_idx[col])
                .min(self.col_idx[col + 1] - 1);
            self.tasks.swap(item, new_idx);
            self.selected_item = new_idx;
        } else {
            match op {
                Op::Decrement => {
                    if self.selected_col > 0 {
                        let task_copy = self.tasks[item].clone();
                        self.remove_task(col, item);
                        self.insert_task(col - 1, task_copy);
                        self.selected_col -= 1;
                    }
                }
                Op::Increment => {
                    if self.selected_col < self.columns - 1 {
                        let task_copy = self.tasks[item].clone();
                        self.remove_task(col, item);
                        self.insert_task(col + 1, task_copy);
                        self.selected_col += 1;
                    }
                }
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    pub fn test_highlighted_item() {
        let mut app = ApplicationState::new(3);
        app.insert_task(app.selected_col, Task::new_test_task());
        app.insert_task(app.selected_col, Task::new_test_task());
        app.insert_task(app.selected_col, Task::new_test_task());
        app.update_selected_column(&Op::Increment);
        app.insert_task(app.selected_col, Task::new_test_task());
        assert_eq!(app.selected_item, 3);

        let sol = app.get_col_slice(1);
        assert!(sol[0].1)
    }

    #[test]
    pub fn test_move_task() {
        let mut app = ApplicationState::new(3);
        app.insert_task(app.selected_col, Task::new_test_task());
        app.insert_task(app.selected_col, Task::new_test_task());
        app.insert_task(app.selected_col, Task::new_test_task());
        app.move_task(app.selected_col, app.selected_item, &Op::Decrement, true);
        assert_eq!(app.selected_item, 1);
        app.move_task(app.selected_col, app.selected_item, &Op::Increment, true);
        assert_eq!(app.selected_item, 2);
    }
}
